@@ -187,11 +187,11 @@ export class ChatInputManager {
   /**
    * Load the user's message history from all chat threads
    */
-  public loadUserMessageHistory(): void {
+  public async loadUserMessageHistory(): Promise<void> {
     this.userMessageHistory = [];
 
     // Iterate through all notebooks
-    const notebookIds = this.chatHistoryManager.getNotebookIds();
+    const notebookIds = await this.chatHistoryManager.getNotebookIds();
     for (const notebookId of notebookIds) {
       // Get all threads for this notebook
       const threads = this.chatHistoryManager.getThreadsForNotebook(notebookId);
