@@ -179,6 +179,7 @@ export class NotebookCellTools {
     position?: number | null;
     show_diff?: boolean;
     tracking_id?: string; // Optional tracking ID to reuse
+    run_cell?: boolean;
   }): string {
     // Get the notebook - pass options.notebook_path directly
     const notebook = this.notebookTools.getCurrentNotebook(
@@ -266,6 +267,10 @@ export class NotebookCellTools {
       newCell.model.sharedModel.setSource(content);
     }
 
+    if (options.run_cell) {
+      void NotebookActions.runCells(notebook, [newCell]);
+    }
+
     // Return the tracking ID instead of cell.model.id since it's more stable
     const trackingMetadata: any =
       newCell.model.sharedModel.getMetadata().cell_tracker;
@@ -360,6 +365,7 @@ export class NotebookCellTools {
     cell_ids: string[];
     notebook_path?: string | null;
     remove_from_notebook?: boolean;
+    save_checkpoint?: boolean;
   }): boolean {
     // Determine if the IDs are tracking IDs
     const useTrackingIds = options.cell_ids.every(id => id.startsWith('cell_'));
@@ -393,15 +399,13 @@ export class NotebookCellTools {
           continue;
         }
 
-        // Get the cell content before removal to show in diff
-        const oldContent = cellInfo.cell.model.sharedModel.getSource();
-
         // Preserve tracking ID for future reference
         // Type-safe access with explicit cast for cell_tracker and its properties
         const cellTracker = metadata.cell_tracker as
           | { trackingId?: string }
           | undefined;
         const trackingId = cellTracker?.trackingId;
+        const oldContent = metadata.custom?.diff?.originalContent;
 
         if (options.remove_from_notebook) {
           // Activate the cell to remove
@@ -412,6 +416,18 @@ export class NotebookCellTools {
           console.log(
             `[NotebookCellTools] Cell ${trackingId} removed permanently from the notebook`
           );
+
+          if (options.save_checkpoint) {
+            AppStateService.getState().actionHistory?.addActionWithCheckpoint(
+              ActionType.REMOVE_CELLS,
+              {
+                cellId,
+                oldContent,
+                metadata
+              },
+              `Removed cell ${cellId.substring(0, 8)}...`
+            );
+          }
         }
 
         cellsRemoved++;
@@ -711,7 +727,9 @@ Generate an updated plan based on this context.`;
           // Handle smart scrolling to show the updated content
           this.handleSmartScroll(planCell.node);
         },
-        { maxTokens: 2048 }
+        { maxTokens: 2048 },
+        undefined,
+        'chat'
       );
 
       // Final content update (in case there are any remaining chunks)
