@@ -1,3 +1,4 @@
+import { PageConfig } from '@jupyterlab/coreutils';
 import Anthropic from '@anthropic-ai/sdk';
 import { ChatRequestStatus } from '../types';
 import { ConfigService } from '../Config/ConfigService';
@@ -7,6 +8,8 @@ import { AnthropicMessageCreator } from './AnthropicMessageCreator';
 import { ServiceUtils } from './ServiceUtils';
 import { AppStateService } from '../AppState';
 
+type Headers = Record<string, string>;
+
 /**
  * Service for handling Anthropic API interactions
  */
@@ -186,6 +189,12 @@ export class AnthropicService implements IChatService {
     await this.validateClientInitialized(errorLogger);
     this.initializeRequest();
 
+    // Always auto-populate custom headers
+    const effectiveCustomHeaders = this.getCustomHeaders({
+      feature: 'chat',
+      mode
+    });
+
     const filteredHistory = await this.getFilteredHistory(errorLogger);
 
     try {
@@ -199,7 +208,8 @@ export class AnthropicService implements IChatService {
         onToolUse,
         notebookContextManager,
         notebookId,
-        errorLogger
+        errorLogger,
+        effectiveCustomHeaders
       );
 
       if (!this.isRequestCancelled()) {
@@ -221,7 +231,8 @@ export class AnthropicService implements IChatService {
         notebookContextManager,
         notebookId,
         errorLogger,
-        forceRetry
+        forceRetry,
+        effectiveCustomHeaders
       );
     }
   }
@@ -247,7 +258,8 @@ export class AnthropicService implements IChatService {
       temperature?: number;
       stopSequences?: string[];
     },
-    onRetry?: (error: Error, attemptNumber: number) => Promise<void>
+    onRetry?: (error: Error, attemptNumber: number) => Promise<void>,
+    feature?: 'cmd-k' | 'chat'
   ): Promise<string> {
     if (!this.client) {
       throw new Error(
@@ -255,13 +267,17 @@ export class AnthropicService implements IChatService {
       );
     }
 
+    // Always auto-populate custom headers
+    const effectiveCustomHeaders = this.getCustomHeaders({ feature });
+
     try {
       return await this.executeEphemeralRequest(
         message,
         systemPrompt,
         modelName,
         onTextChunk,
-        options
+        options,
+        effectiveCustomHeaders
       );
     } catch (error: any) {
       return await this.handleEphemeralRequestError(
@@ -271,7 +287,8 @@ export class AnthropicService implements IChatService {
         modelName,
         onTextChunk,
         options,
-        onRetry
+        onRetry,
+        effectiveCustomHeaders
       );
     }
   }
@@ -288,26 +305,32 @@ export class AnthropicService implements IChatService {
       maxTokens?: number;
       temperature?: number;
       stopSequences?: string[];
-    }
+    },
+    headers: Headers = {}
   ): Promise<string> {
     const maxTokens = options?.maxTokens ?? 48;
     const temperature = options?.temperature ?? 0;
 
     if (onTextChunk) {
       // Use streaming for real-time updates
-      const stream = await this.client!.messages.stream({
-        model: modelName,
-        max_tokens: maxTokens,
-        system: systemPrompt,
-        temperature,
-        stop_sequences: options?.stopSequences,
-        messages: [
-          {
-            role: 'user',
-            content: message
-          }
-        ]
-      });
+      const stream = await this.client!.messages.stream(
+        {
+          model: modelName,
+          max_tokens: maxTokens,
+          system: systemPrompt,
+          temperature,
+          stop_sequences: options?.stopSequences,
+          messages: [
+            {
+              role: 'user',
+              content: message
+            }
+          ]
+        },
+        {
+          headers: Object.keys(headers).length > 0 ? headers : undefined
+        }
+      );
 
       let completeResponse = '';
 
@@ -325,19 +348,24 @@ export class AnthropicService implements IChatService {
       return completeResponse;
     } else {
       // Use non-streaming for simple responses
-      const response = await this.client!.messages.create({
-        model: modelName,
-        max_tokens: maxTokens,
-        system: systemPrompt,
-        temperature,
-        stop_sequences: options?.stopSequences,
-        messages: [
-          {
-            role: 'user',
-            content: message
-          }
-        ]
-      });
+      const response = await this.client!.messages.create(
+        {
+          model: modelName,
+          max_tokens: maxTokens,
+          system: systemPrompt,
+          temperature,
+          stop_sequences: options?.stopSequences,
+          messages: [
+            {
+              role: 'user',
+              content: message
+            }
+          ]
+        },
+        {
+          headers: Object.keys(headers).length > 0 ? headers : undefined
+        }
+      );
 
       // Extract text content from response
       let responseText = '';
@@ -365,7 +393,8 @@ export class AnthropicService implements IChatService {
       temperature?: number;
       stopSequences?: string[];
     },
-    onRetry?: (error: Error, attemptNumber: number) => Promise<void>
+    onRetry?: (error: Error, attemptNumber: number) => Promise<void>,
+    customHeaders: Headers = {}
   ): Promise<string> {
     console.error('Ephemeral message error:', error);
 
@@ -391,7 +420,8 @@ export class AnthropicService implements IChatService {
           systemPrompt,
           modelName,
           onTextChunk,
-          options
+          options,
+          customHeaders
         );
       } catch (retryError: any) {
         // If this is the last attempt, throw the error
@@ -487,7 +517,8 @@ export class AnthropicService implements IChatService {
     onToolUse?: (toolUse: any) => void,
     notebookContextManager?: any,
     notebookId?: string,
-    errorLogger?: (message: any) => Promise<void>
+    errorLogger?: (message: any) => Promise<void>,
+    customHeaders?: Headers
   ): Promise<any> {
     const stream = await this.createMessageStream(
       filteredHistory,
@@ -497,7 +528,8 @@ export class AnthropicService implements IChatService {
       fetchNotebookState,
       notebookContextManager,
       notebookId,
-      errorLogger
+      errorLogger,
+      customHeaders
     );
 
     return await AnthropicStreamHandler.processStream(
@@ -523,7 +555,8 @@ export class AnthropicService implements IChatService {
     fetchNotebookState?: () => Promise<string>,
     notebookContextManager?: any,
     notebookId?: string,
-    errorLogger?: (message: any) => Promise<void>
+    errorLogger?: (message: any) => Promise<void>,
+    customHeaders: Headers = {}
   ): Promise<any> {
     try {
       return await AnthropicMessageCreator.createMessageStream(
@@ -542,7 +575,8 @@ export class AnthropicService implements IChatService {
           notebookContextManager,
           notebookId: notebookId,
           abortSignal: this.abortController!.signal,
-          errorLogger
+          errorLogger,
+          customHeaders
         },
         filteredHistory,
         ServiceUtils.normalizeMessageContent
@@ -579,7 +613,14 @@ export class AnthropicService implements IChatService {
     notebookContextManager?: any,
     notebookPath?: string,
     errorLogger?: (message: any) => Promise<void>,
-    forceRetry: boolean = false
+    forceRetry: boolean = false,
+    customHeaders: {
+      notebookId?: string;
+      clientExtVersion?: string;
+      clientLabVersion?: string;
+      sageSessionId?: string;
+      sageThreadId?: string;
+    } = {}
   ): Promise<any> {
     if (this.isRequestCancelled()) {
       return ServiceUtils.createCancelledResponse(
@@ -607,7 +648,8 @@ export class AnthropicService implements IChatService {
         onToolUse,
         notebookContextManager,
         notebookPath,
-        errorLogger
+        errorLogger,
+        customHeaders
       );
     }
 
@@ -674,7 +716,14 @@ export class AnthropicService implements IChatService {
     onToolUse?: (toolUse: any) => void,
     notebookContextManager?: any,
     notebookPath?: string,
-    errorLogger?: (message: any) => Promise<void>
+    errorLogger?: (message: any) => Promise<void>,
+    customHeaders: {
+      notebookId?: string;
+      clientExtVersion?: string;
+      clientLabVersion?: string;
+      sageSessionId?: string;
+      sageThreadId?: string;
+    } = {}
   ): Promise<any> {
     this.requestStatus = ChatRequestStatus.RETRYING;
 
@@ -714,7 +763,8 @@ export class AnthropicService implements IChatService {
           onToolUse,
           notebookContextManager,
           notebookPath,
-          errorLogger
+          errorLogger,
+          customHeaders
         );
 
         if (!this.isRequestCancelled()) {
@@ -806,4 +856,72 @@ export class AnthropicService implements IChatService {
   setToolBlacklist(blacklist: string[]): void {
     this.toolBlacklist = Array.isArray(blacklist) ? [...blacklist] : [];
   }
+
+  /**
+   * Gets custom headers by auto-populating from application state
+   * @returns Headers object with auto-populated values where available
+   */
+  private getCustomHeaders({
+    feature,
+    mode
+  }: {
+    feature?: 'cmd-k' | 'chat';
+    mode?: 'agent' | 'ask' | 'fast';
+  }): Headers {
+    const headers: {
+      'X-Sage-Mode'?: 'agent' | 'ask' | 'fast';
+      'X-Sage-Feature'?: 'cmd-k' | 'chat';
+      'X-Sage-Notebook-Id'?: string;
+      'X-Sage-Client-Ext-Version'?: string;
+      'X-Sage-Client-Lab-Version'?: string;
+      'X-Sage-Session-Id'?: string;
+      'X-Sage-Thread-Id'?: string;
+    } = {};
+
+    if (mode) {
+      headers['X-Sage-Mode'] = mode;
+    }
+
+    if (feature) {
+      headers['X-Sage-Feature'] = feature;
+    }
+
+    // Always get notebook ID from AppState
+    const notebookId = AppStateService.getCurrentNotebookId();
+    if (notebookId) {
+      headers['X-Sage-Notebook-Id'] = notebookId;
+    }
+
+    // Always get session ID from notebook tracker if available
+    const notebookTracker = AppStateService.getState().notebookTracker;
+    const currentWidget = notebookTracker?.currentWidget;
+    const sageSessionId = currentWidget?.sessionContext?.session?.id;
+    if (sageSessionId) {
+      headers['X-Sage-Session-Id'] = sageSessionId;
+    }
+
+    const threadId =
+      AppStateService.getState().chatContainer?.chatWidget.chatHistoryManager.getCurrentThread()
+        ?.id;
+    if (threadId) {
+      headers['X-Sage-Thread-Id'] = threadId;
+    }
+
+    const labVersion = PageConfig.getOption('appVersion');
+    if (labVersion) {
+      headers['X-Sage-Client-Lab-Version'] = labVersion;
+    }
+
+    const extensions = AppStateService.getExtensions();
+    if (extensions) {
+      const sageExtension = extensions.installed.find(
+        value => value.name === 'sage-agent-internal'
+      );
+      if (sageExtension) {
+        headers['X-Sage-Client-Ext-Version'] = sageExtension.installed_version;
+      }
+    }
+
+    return headers;
+  }
 }
