@@ -95,7 +95,8 @@ export interface IChatService {
       temperature?: number;
       stopSequences?: string[];
     },
-    onRetry?: (error: Error, attemptNumber: number) => Promise<void>
+    onRetry?: (error: Error, attemptNumber: number) => Promise<void>,
+    feature?: 'cmd-k' | 'chat'
   ): Promise<string>;
 
   /**
